/** BEGIN COPYRIGHT BLOCK
 * This Program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; version 2 of the License.
 *
 * This Program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this Program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * In addition, as a special exception, Red Hat, Inc. gives You the additional
 * right to link the code of this Program with code not covered under the GNU
 * General Public License ("Non-GPL Code") and to distribute linked combinations
 * including the two, subject to the limitations in this paragraph. Non-GPL Code
 * permitted under this exception must only link to the code of this Program
 * through those well defined interfaces identified in the file named EXCEPTION
 * found in the source code files (the "Approved Interfaces"). The files of
 * Non-GPL Code may instantiate templates or use macros or inline functions from
 * the Approved Interfaces without causing the resulting work to be covered by
 * the GNU General Public License. Only Red Hat, Inc. may make changes or
 * additions to the list of Approved Interfaces. You must obey the GNU General
 * Public License in all respects for all of the Program code and other code used
 * in conjunction with the Program except the Non-GPL Code covered by this
 * exception. If you modify this file, you may extend this exception to your
 * version of the file, but you are not obligated to do so. If you do not wish to
 * provide this exception without modification, you must delete this exception
 * statement from your version and license this file solely under the GPL without
 * exception.
 *
 *
 * Copyright (C) 2013 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <libotp.h>
#include <time.h>

#define PLUGIN_NAME               "ipa-otp-lasttoken"
#define LOG(sev, ...) \
    slapi_log_error(SLAPI_LOG_ ## sev, PLUGIN_NAME, \
                    "%s: %s\n", __func__, __VA_ARGS__), -1

static void *plugin_id;
static const Slapi_PluginDesc preop_desc = {
    PLUGIN_NAME,
    "FreeIPA",
    "FreeIPA/1.0",
    "Protect the user's last active token"
};

static bool
target_is_only_enabled_token(Slapi_PBlock *pb)
{
    Slapi_DN *target_sdn = NULL;
    Slapi_DN *token_sdn = NULL;
    Slapi_DN *owner_sdn = NULL;
    struct otptoken **tokens;
    char *dn = NULL;
    bool match;

    /* Ignore internal operations. */
    if (slapi_op_internal(pb))
        return false;

    /* Get the current user's SDN. */
    slapi_pblock_get(pb, SLAPI_CONN_DN, &dn);
    if (dn == NULL)
        return false;
    owner_sdn = slapi_sdn_new_dn_passin(dn);
    if (owner_sdn == NULL) {
        slapi_ch_free_string(&dn);
        return false;
    }

    /* Get the SDN of the only enabled token. */
    tokens = otptoken_find(plugin_id, owner_sdn, NULL, true, NULL);
    slapi_sdn_free(&owner_sdn);
    if (tokens != NULL && tokens[0] != NULL && tokens[1] == NULL)
        token_sdn = slapi_sdn_dup(otptoken_get_sdn(tokens[0]));
    otptoken_free_array(tokens);
    if (token_sdn == NULL)
        return false;

    /* Get the target SDN. */
    slapi_pblock_get(pb, SLAPI_TARGET_SDN, &target_sdn);
    if (target_sdn == NULL) {
        slapi_sdn_free(&token_sdn);
        return false;
    }

    /* Does the target SDN match the only enabled token SDN? */
    match = slapi_sdn_compare(token_sdn, target_sdn) == 0;
    slapi_sdn_free(&token_sdn);
    return match;
}

static inline int
send_error(Slapi_PBlock *pb, int rc, char *errstr)
{
    slapi_send_ldap_result(pb, rc, NULL, errstr, 0, NULL);
    slapi_pblock_set(pb, SLAPI_RESULT_CODE, &rc);
    return rc;
}

static int
preop_del(Slapi_PBlock *pb)
{
    if (!target_is_only_enabled_token(pb))
        return 0;

    return send_error(pb, LDAP_UNWILLING_TO_PERFORM,
                      "Can't delete last active token");
}

static int
preop_mod(Slapi_PBlock *pb)
{
    LDAPMod **mods = NULL;

    if (!target_is_only_enabled_token(pb))
        return 0;

    /* Do not permit deactivation of the last active token. */
    slapi_pblock_get(pb, SLAPI_MODIFY_MODS, &mods);
    for (int i = 0; mods != NULL && mods[i] != NULL; i++) {
        if (strcasecmp(mods[i]->mod_type, "ipatokenDisabled") == 0) {
            return send_error(pb, LDAP_UNWILLING_TO_PERFORM,
                              "Can't disable last active token");
        }

        if (strcasecmp(mods[i]->mod_type, "ipatokenOwner") == 0) {
            return send_error(pb, LDAP_UNWILLING_TO_PERFORM,
                              "Can't change last active token's owner");
        }

        if (strcasecmp(mods[i]->mod_type, "ipatokenNotBefore") == 0) {
            return send_error(pb, LDAP_UNWILLING_TO_PERFORM,
                              "Can't change last active token's start time");
        }

        if (strcasecmp(mods[i]->mod_type, "ipatokenNotAfter") == 0) {
            return send_error(pb, LDAP_UNWILLING_TO_PERFORM,
                              "Can't change last active token's end time");
        }
    }

    return 0;
}

static int
preop_init(Slapi_PBlock *pb)
{
    if (slapi_pblock_set(pb, SLAPI_PLUGIN_VERSION, SLAPI_PLUGIN_VERSION_01))
        goto error;

    if (slapi_pblock_set(pb, SLAPI_PLUGIN_DESCRIPTION, (void *) &preop_desc))
        goto error;

    if (slapi_pblock_set(pb, SLAPI_PLUGIN_BE_TXN_PRE_DELETE_FN, preop_del))
        goto error;

    if (slapi_pblock_set(pb, SLAPI_PLUGIN_BE_TXN_PRE_MODIFY_FN, preop_mod))
        goto error;

    return 0;

error:
    return LOG(FATAL, "failed to register be_txn_pre_op plugin");
}

int
ipa_otp_lasttoken_init(Slapi_PBlock *pb)
{
    slapi_pblock_get(pb, SLAPI_PLUGIN_IDENTITY, &plugin_id);

    if (slapi_register_plugin("betxnpreoperation", 1, __func__, preop_init,
                              PLUGIN_NAME, NULL, plugin_id))
        return LOG(FATAL, "failed to register plugin");

    return 0;
}
