/** BEGIN COPYRIGHT BLOCK
 * This Program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; version 2 of the License.
 *
 * This Program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this Program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * In addition, as a special exception, Red Hat, Inc. gives You the additional
 * right to link the code of this Program with code not covered under the GNU
 * General Public License ("Non-GPL Code") and to distribute linked combinations
 * including the two, subject to the limitations in this paragraph. Non-GPL Code
 * permitted under this exception must only link to the code of this Program
 * through those well defined interfaces identified in the file named EXCEPTION
 * found in the source code files (the "Approved Interfaces"). The files of
 * Non-GPL Code may instantiate templates or use macros or inline functions from
 * the Approved Interfaces without causing the resulting work to be covered by
 * the GNU General Public License. Only Red Hat, Inc. may make changes or
 * additions to the list of Approved Interfaces. You must obey the GNU General
 * Public License in all respects for all of the Program code and other code used
 * in conjunction with the Program except the Non-GPL Code covered by this
 * exception. If you modify this file, you may extend this exception to your
 * version of the file, but you are not obligated to do so. If you do not wish to
 * provide this exception without modification, you must delete this exception
 * statement from your version and license this file solely under the GPL without
 * exception.
 *
 *
 * Copyright (C) 2013 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "request.h"

#define PLUGIN_NAME               "ipa-otp-sync"
#define PLUGIN_DESC               "otp sync extended operation plugin"

static void *plugin_id;

static const Slapi_PluginDesc extop_desc = {
    PLUGIN_NAME,
    "FreeIPA",
    "FreeIPA/1.0",
    "Synchronize a user's token"
};

static const char const *oid_list[] = {
    OTP_SYNC_REQUEST_OID,
    NULL
};

static const char const *name_list[] = {
    "otp_sync_extop",
    NULL
};

/*
 * Extended operation function:
 *   1. Parse the request.
 *   2. Authenticate the connection via the bind or the credentials provided.
 *   3. Synchronize the token(s).
 *   4. Send the result.
 */
static int
otp_sync_extop(Slapi_PBlock *pb)
{
    struct otp_sync_request *req = NULL;
    char *error = NULL;
    int rc = 0;

    /* Parse the request. */
    req = otp_sync_request_new(plugin_id, pb);
    if (req == NULL) {
        error = "Error decoding request";
        rc = LDAP_OPERATIONS_ERROR;
        goto egress;
    }

    /* Perform authentication. */
    if (!otp_sync_request_authenticate(req)) {
        error = "Authentication failed";
        rc = LDAP_INVALID_CREDENTIALS;
        goto egress;
    }

    /* Perform the sync. */
    if (!otp_sync_request_sync(req)) {
        error = "Token synchronization failed";
        rc = LDAP_OPERATIONS_ERROR;
        goto egress;
    }

egress:
    if (error != NULL)
        slapi_log_error(SLAPI_LOG_PLUGIN, PLUGIN_NAME, "%s\n", error);
    slapi_send_ldap_result(pb, rc, NULL, error, 0, NULL);
    slapi_pblock_set(pb, SLAPI_RESULT_CODE, &rc);
    otp_sync_request_free(req);
    return SLAPI_PLUGIN_EXTENDED_SENT_RESULT;
}

static int
otp_sync_extop_init(Slapi_PBlock *pb)
{
    if (slapi_pblock_set(pb, SLAPI_PLUGIN_VERSION, SLAPI_PLUGIN_VERSION_01))
        goto error;

    if (slapi_pblock_set(pb, SLAPI_PLUGIN_DESCRIPTION, (void *) &extop_desc))
        goto error;

    if (slapi_pblock_set(pb, SLAPI_PLUGIN_EXT_OP_OIDLIST, oid_list))
        goto error;

    if (slapi_pblock_set(pb, SLAPI_PLUGIN_EXT_OP_NAMELIST, name_list))
        goto error;

    if (slapi_pblock_set(pb, SLAPI_PLUGIN_EXT_OP_FN, otp_sync_extop))
        goto error;

    return 0;

error:
    slapi_log_error(SLAPI_LOG_FATAL, PLUGIN_NAME,
                    "%s: failed to register extended operation plugin\n",
                    __func__);
    return -1;
}

int
ipa_otp_sync_init(Slapi_PBlock *pb)
{
    slapi_pblock_get(pb, SLAPI_PLUGIN_IDENTITY, &plugin_id);

    if (slapi_register_plugin("extendedop", 1, __func__, otp_sync_extop_init,
                              PLUGIN_DESC, NULL, plugin_id)) {
        slapi_log_error(SLAPI_LOG_FATAL, PLUGIN_NAME,
                        "%s: failed to register plugin\n", __func__);
        return -1;
    }

    return 0;
}
