/** BEGIN COPYRIGHT BLOCK
 * This Program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; version 2 of the License.
 *
 * This Program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this Program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * In addition, as a special exception, Red Hat, Inc. gives You the additional
 * right to link the code of this Program with code not covered under the GNU
 * General Public License ("Non-GPL Code") and to distribute linked combinations
 * including the two, subject to the limitations in this paragraph. Non-GPL Code
 * permitted under this exception must only link to the code of this Program
 * through those well defined interfaces identified in the file named EXCEPTION
 * found in the source code files (the "Approved Interfaces"). The files of
 * Non-GPL Code may instantiate templates or use macros or inline functions from
 * the Approved Interfaces without causing the resulting work to be covered by
 * the GNU General Public License. Only Red Hat, Inc. may make changes or
 * additions to the list of Approved Interfaces. You must obey the GNU General
 * Public License in all respects for all of the Program code and other code used
 * in conjunction with the Program except the Non-GPL Code covered by this
 * exception. If you modify this file, you may extend this exception to your
 * version of the file, but you are not obligated to do so. If you do not wish to
 * provide this exception without modification, you must delete this exception
 * statement from your version and license this file solely under the GPL without
 * exception.
 *
 *
 * Copyright (C) 2013 Red Hat, Inc.
 * All rights reserved.
 * END COPYRIGHT BLOCK **/

#include "request.h"
#include <libotp.h>

#define OTP_SYNC_TAG_TOKEN_DN     0x80U
#define OTP_SYNC_TAG_FIRST_FACTOR 0x81U
#define OTP_SYNC_MAX_STEPS        25

struct otp_sync_request {
    void *plugin_id;
    Slapi_PBlock *request;
    Slapi_DN *userDN;
    Slapi_DN *tokenDN;
    char *firstFactor;
    int firstCode;
    int secondCode;
};

void otp_sync_request_free(struct otp_sync_request *req)
{
    if (req == NULL)
        return;

    slapi_sdn_free(&req->userDN);
    slapi_sdn_free(&req->tokenDN);
    slapi_ch_free_string(&req->firstFactor);
    free(req);
}

struct otp_sync_request *otp_sync_request_new(void *plugin_id,
                                              Slapi_PBlock *pb)
{
    struct otp_sync_request *req = NULL;
    ber_len_t len = (ber_len_t) -1;
    struct berval *reqdata = NULL;
    BerElement *ber = NULL;
    char *token_dn = NULL;
    char *user_dn = NULL;
    char *oid = NULL;

    /* Create the request object. */
    req = calloc(1, sizeof(*req));
    if (!req)
        return NULL;
    req->plugin_id = plugin_id;
    req->request = pb;

    /* Get the OID and the value included in the request */
    if (slapi_pblock_get(pb, SLAPI_EXT_OP_REQ_OID, &oid) != 0)
        goto error;

    /* Make sure the request OID is correct. */
    if (oid == NULL || strcmp(oid, OTP_SYNC_REQUEST_OID) != 0)
        goto error;

    /* Get the request value */
    if (slapi_pblock_get(pb, SLAPI_EXT_OP_REQ_VALUE, &reqdata) != 0)
        goto error;
    if (!BV_HAS_DATA(reqdata))
        goto error;

    /* Decode the request. */
    ber = ber_init(reqdata);
    if (ber == NULL)
        goto error;

    /* Get the user's DN. */
    if (ber_scanf(ber, "{a", &user_dn) == LBER_ERROR)
        goto error;
    req->userDN = slapi_sdn_new_dn_passin(user_dn);
    if (req->userDN == NULL) {
        slapi_ch_free_string(&user_dn);
        goto error;
    }

    /* Get the token's DN (optional). */
    if (ber_peek_tag(ber, &len) == OTP_SYNC_TAG_TOKEN_DN) {
        if (ber_scanf(ber, "a", &token_dn) == LBER_ERROR)
            goto error;
        req->tokenDN = slapi_sdn_new_dn_passin(token_dn);
        if (req->tokenDN == NULL) {
            slapi_ch_free_string(&token_dn);
            goto error;
        }
    }

    /* Get the user's password (optional). */
    if (ber_peek_tag(ber, &len) == OTP_SYNC_TAG_FIRST_FACTOR) {
        if (ber_scanf(ber, "a", &req->firstFactor) == LBER_ERROR)
            goto error;
    }

    /* Get the first and second token code. */
    if (ber_scanf(ber, "i", &req->firstCode) == LBER_ERROR)
        goto error;
    if (ber_scanf(ber, "i}", &req->secondCode) == LBER_ERROR)
        goto error;

    ber_free(ber, 1);
    return req;

error:
    if (NULL != ber)
        ber_free(ber, 1);
    otp_sync_request_free(req);
    return NULL;
}

bool otp_sync_request_authenticate(struct otp_sync_request *req)
{
    Slapi_PBlock *search_pb = NULL;
    Slapi_Entry **entries = NULL;
    char *bind_dn_str = NULL;
    PWScheme *scheme = NULL;
    char *password = NULL;
    bool success = false;
    char *attr = NULL;
    int result = 0;

    /* If there is a bind, just ensure the DNs are the same. */
    if (slapi_pblock_get(req->request, SLAPI_CONN_DN, &bind_dn_str) == 0 &&
        bind_dn_str != NULL) {
        Slapi_DN *bind_dn;
        PRBool match;

        bind_dn = slapi_sdn_new_dn_passin(bind_dn_str);
        match = slapi_sdn_compare(bind_dn, req->userDN) == 0;
        slapi_sdn_free(&bind_dn);
        return match;
    }

    /* If firstFactor is missing, we can't authenticate. */
    if (req->firstFactor == NULL)
        return false;

    /* Get the user's entry. */
    search_pb = slapi_pblock_new();
    slapi_search_internal_set_pb(search_pb, slapi_sdn_get_dn(req->userDN),
                                 LDAP_SCOPE_BASE, "objectclass=*",
                                 (char*[]) { "userPassword", NULL },
                                 1, NULL, NULL, req->plugin_id, 0);
    slapi_search_internal_pb(search_pb);
    if (slapi_pblock_get(search_pb, SLAPI_PLUGIN_INTOP_RESULT, &result)
        || result != LDAP_SUCCESS
        || slapi_pblock_get(search_pb, SLAPI_PLUGIN_INTOP_SEARCH_ENTRIES, &entries)
        || entries == NULL
        || entries[0] == NULL)
        goto error;

    /* Get the user's password. */
    attr = slapi_entry_attr_get_charptr(entries[0], "userPassword");
    if (attr == NULL)
        goto error;

    /* Compare the passwords. */
    scheme = slapi_pw_val2scheme(attr, &password, 1);
    if (scheme == NULL)
        goto error;
    if (strcmp(slapi_pw_get_scheme_name(scheme), "CLEAR") == 0) {
        /* Password is in clear text. */
        if (strcmp(req->firstFactor, password) != 0)
            goto error;
    } else {
        /* Password is encoded. */
        if (slapi_pw_cmp(scheme, req->firstFactor, password) != 0)
            goto error;
    }

    success = true;

error:
    slapi_free_search_results_internal(search_pb);
    slapi_pblock_destroy(search_pb);
    slapi_free_pw_scheme(scheme);
    slapi_ch_free_string(&attr);
    return success;
}

bool otp_sync_request_sync(struct otp_sync_request *req)
{
    struct otptoken **tokens = NULL;
    bool success = false;

    tokens = otptoken_find(req->plugin_id, req->userDN, req->tokenDN, true, NULL);
    if (tokens != NULL)
        success = otptoken_sync(tokens, OTP_SYNC_MAX_STEPS,
                                req->firstCode, req->secondCode);

    otptoken_free_array(tokens);
    return success;
}
